#include "main.h"
#include "extern.h"

//==========================AVR_7SEG_MODBUS_CODE====================================================
//==============================================================================
//=========================CRC GENRATION CODE ==================================
//==============================================================================
#define READ_ONLY			0x04
#define READ_HOLDING		0x03
//#define WRITE_SINGLE		0x06

#define READ_ONLY_MAXADDR   3000
#define READ_ONLY_MINADDR   3000

#define HOLDING_MAXADDR     4085
#define HOLDING_MINADDR     4000

#define ADDRESS_ERROR		0x02 
#define FUNCTION_ERROR		0x01 

#define ERROR				1
#define OK					0

//#define TAPADDR				3000
//#define PTVOLTADRR			3001
//==============================================================================

//==============================================================================
unsigned int CRC_16(unsigned char *ptr,unsigned char cnts)
{
	union int_char crcreg;
	unsigned char i;
	crcreg.int_val = 0xffff;
    while(cnts !=0)
	{
		crcreg.char_val[LOWBYTE] ^= *ptr;
		for(i=8; i!=0; i--)
		{
			if(crcreg.char_val[LOWBYTE] & 0x01 == 1)
			{	crcreg.int_val = crcreg.int_val>>1;
				crcreg.int_val ^= 0xa001;
			}
			else
			{
                crcreg.int_val= crcreg.int_val>>1;
			}
		}

		cnts--;
		ptr++;
	}
	return(crcreg.int_val);
}
//==============================================================================
//===================DECODE RECIEVED FRAME =====================================
//==============================================================================
/* FUNCTION:decode_mbus_frame()
 * this function is call if all mbus packet is received 
 * and receive_frame_cmplt   flag is set 
 * This function check function code 
 * address and length of packet demand 
    and process according to
 * function code 
 * 
 */
//==============================================================================

void decode_mbus_frame(void)
{
    
    if(mbus_rx_framecmplt_flg==0)return;
	
    union int_char un1;
    unsigned int address,length;
    unsigned int temp;
    
    
    mbus_rx_framestart_flg=0;
    mbus_rx_framecmplt_flg=0;
    mbus_receive_tmr=0;
	
    if(slave_id != mbus_rx_buff[0])
    {
        mbus_ptr=0;
	    return;
    }
	
    //check crc
    temp= CRC_16(mbus_rx_buff,mbus_ptr-2);       //mbus -2 defines legth of data
    un1.char_val[HIGHBYTE]=mbus_rx_buff[mbus_ptr-1];
    un1.char_val[LOWBYTE]=mbus_rx_buff[mbus_ptr-2];
	
    if(un1.int_val!=temp)
    {
        mbus_ptr=0;
		return;                 //if crc is match return;
    }
    //   0    ,  0     ,   0 , 0       ,   0 , 0 
   // slave id, function , starting addr, length // 0 
    un1.char_val[LOWBYTE]=mbus_rx_buff[3];
    un1.char_val[HIGHBYTE]=mbus_rx_buff[2];
    address=un1.int_val;                                          //read address
	
    un1.char_val[LOWBYTE]=mbus_rx_buff[5];
    un1.char_val[HIGHBYTE]=mbus_rx_buff[4];
    length=un1.int_val;                                            //read length
	
    //check command  and take action  .........................................
    switch(mbus_rx_buff[1])                                      
    {
		case READ_HOLDING	:	
                                if(check_addr_error_holding_register(address,length) == 0)
								{
									fill_holding_register(address,length);
								}
								else
								{
									address_error(mbus_rx_buff[1]);
								}
        break;
        
        case READ_ONLY		:   
            
                                if(check_addr_error_input_register(address,length) == 0)
								{
                                    fill_readonly_byte(address,length);
								}
								else
								{
									address_error(mbus_rx_buff[1]);
								}
       break;
       default				:   
                                function_error(mbus_rx_buff[1]);
       break;            
                
    }
    
    mbus_ptr=0;
	
}
//==============================================================================
/*
    function:function_error
    call when recive buffer have function cmd error 
 */
//==============================================================================
void function_error(unsigned int func)
{
    unsigned int temp;
    union int_char un1;
    temp=func;
    mbus_tx_buff[0] = slave_id;                                     //salve id
    mbus_tx_buff[1] = temp|0x80;                                    // make msb  bit high
    mbus_tx_buff[2] = FUNCTION_ERROR;                                         //signal function error              
    un1.int_val = CRC_16(mbus_tx_buff,2);                          
    mbus_tx_buff[3] =  un1.char_val[LOWBYTE];                       //lsb of crc  
    mbus_tx_buff[4] = un1.char_val[HIGHBYTE];                       //msb of crc
    mbus_send_frame_flg=1;                                          //ENABLE SEND _FRAME FLAG
    mbus_send_frame_len=5;                                         
    mbus_send_ptr=0;
    
    mbus_SendFirstByte();
    
}
//==============================================================================
    /*function : address_error
    * call when receive buffer have address  error 
     */ 
//==============================================================================
void address_error(unsigned int func)
{
  unsigned int temp;
  union int_char un1;
  
  temp=func;
  mbus_tx_buff[0] = slave_id;
  mbus_tx_buff[1] = temp | 0x80;
  mbus_tx_buff[2] = ADDRESS_ERROR;			//signal address error
  un1.int_val = CRC_16(mbus_tx_buff,2);
  mbus_tx_buff[3] =  un1.char_val[LOWBYTE];  //lsb of crc  
  mbus_tx_buff[4] = un1.char_val[HIGHBYTE];  //msb of crc
  
  mbus_send_frame_flg = 1;
  
  mbus_send_frame_len = 5;
  
  mbus_send_ptr=0;
  
  mbus_SendFirstByte();                 //sid
  
}
//==============================================================================
/*
    FUNCTION:fill_nsende_echo_frame
    * CALL AFTER RESPONS OF WRITE REGISTER CMD 
    * SEND  ECHO OF RECEIVE BUFF TO MASTER
 */

//==============================================================================
void FillNSend_Echo_Frame(void)
{
    unsigned char i=0;
    
    for(i=0;i<mbus_ptr;i++)
    {
        mbus_tx_buff[i]=mbus_rx_buff[i];
    }
    mbus_send_frame_len = mbus_ptr;
    mbus_ptr=0;
    mbus_SendFirstByte();
    //sam
}
//==============================================================================
unsigned char check_addr_error_holding_register(unsigned int adrr,unsigned char len)
{
  if((adrr > HOLDING_MAXADDR)||(adrr < HOLDING_MINADDR))
        return(ERROR);
  if((adrr+len) > HOLDING_MAXADDR+1)
		return(ERROR);
  else
	return(OK);
}
//==============================================================================

unsigned char check_addr_error_input_register(unsigned int adrr,unsigned char len)
{

    if((adrr > READ_ONLY_MAXADDR)||(adrr < READ_ONLY_MINADDR))
        return(ERROR);
    if((adrr+len)>READ_ONLY_MAXADDR+1)   // max number of length to be 
		return(ERROR);
    else
		return(OK);
}
//==============================================================================
/*
 * FUNCTION:mbus_send_frame
 * CALL IN ISR 
 * SEND ONE BYTE IN EVERY 1MS 
 * TILL mbus_send_ptr EQUAL TO mbus_send_frame_cmpr
 */
//==============================================================================
/*void mbus_send_frame(void)
{
   if(mbus_send_frame_flg == 0)return;
   if(mbus_send_ptr > mbus_send_frame_len)
	{	
		mbus_send_ptr = 0;
		mbus_send_frame_flg = 0;
        return;
	}
    Uart1sendbyte(mbus_tx_buff[mbus_send_ptr]);
    mbus_send_ptr++;	
}
*/
//==============================================================================
void fill_readonly_byte(unsigned int addr ,unsigned int len)
{
    union int_char un1;
    
    un1.int_val=len;
    
    unsigned int tmp,i;
    
    tmp = addr-0;
    
    mbus_tx_buff[0] = slave_id;
    
    mbus_tx_buff[1] = READ_ONLY;                 
	
    mbus_tx_buff[2] = un1.char_val[LOWBYTE]<<1;		//byte count
    
    mbus_send_frame_len = 3;
        
    un1.int_val=tap_count;
    
    mbus_tx_buff[mbus_send_frame_len] = un1.char_val[HIGHBYTE];//msb of crc
    mbus_send_frame_len++;
    mbus_tx_buff[mbus_send_frame_len] = un1.char_val[LOWBYTE];
    mbus_send_frame_len++;
    
    
    un1.int_val = CRC_16(mbus_tx_buff,mbus_send_frame_len);  
    mbus_tx_buff[mbus_send_frame_len] =  un1.char_val[LOWBYTE];  //lsb of crc  
    mbus_send_frame_len++;
    mbus_tx_buff[mbus_send_frame_len] = un1.char_val[HIGHBYTE];  //msb of crc
    
    mbus_send_frame_len++;
    mbus_send_frame_flg = 1;                                          //ENABLE SEND _FRAME FLAG
    
    mbus_send_ptr=0;
    
    mbus_SendFirstByte();
    
}

 //==============================================================================
void fill_holding_register(unsigned char addr,unsigned char len)
{
    union int_char un1;
    unsigned char e2p_adr;
    unsigned int i=0,tmp;
    unsigned char j=0;
    
    un1.int_val=len;
    
    tmp=addr;             
	
    mbus_tx_buff[0] = slave_id;
    mbus_tx_buff[1] = READ_HOLDING;
    mbus_tx_buff[2] = un1.char_val[LOWBYTE]<<1; //byte count
    
    mbus_send_frame_len=3;
    
    j = addr-HOLDING_MINADDR;    
    j += j;
    
    for(i = tmp; i<(tmp+len); i++)       
    {
        
//        e2p_adr =  (START_READ_HOLDING_ADDR+j);                       
//        j+=2;
//        un1.int_val= i2c_int_rd(e2p_adr);
        mbus_tx_buff[mbus_send_frame_len] = un1.char_val[HIGHBYTE];//msb
        mbus_send_frame_len++;
        mbus_tx_buff[mbus_send_frame_len] = un1.char_val[LOWBYTE];  //lsb  
        mbus_send_frame_len++;
        
    }
    
    
    un1.int_val = CRC_16(mbus_tx_buff,mbus_send_frame_len);  // 5 is fix modify latter ajay 
    
    mbus_tx_buff[mbus_send_frame_len] = un1.char_val[LOWBYTE];//msb of crc
    mbus_send_frame_len++;
    mbus_tx_buff[mbus_send_frame_len] = un1.char_val[HIGHBYTE];
    mbus_send_frame_len++;
    
    mbus_send_frame_flg=1;                                          //ENABLE SEND _FRAME FLAG
    mbus_send_ptr=0;
    
    MAX_485_PIN=1;
   
    mbus_SendFirstByte();          
    
 }

//==============================================================================
//===================================================================================================





//===================================================================================================